/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef __APP_PIKA_UNIT_H__
#define __APP_PIKA_UNIT_H__


gint          _pika_unit_get_number_of_units          (Pika        *pika);
gint          _pika_unit_get_number_of_built_in_units (Pika        *pika) G_GNUC_CONST;

PikaUnit      _pika_unit_new                          (Pika        *pika,
                                                       const gchar *identifier,
                                                       gdouble      factor,
                                                       gint         digits,
                                                       const gchar *symbol,
                                                       const gchar *abbreviation,
                                                       const gchar *singular,
                                                       const gchar *plural);

gboolean      _pika_unit_get_deletion_flag            (Pika        *pika,
                                                       PikaUnit     unit);
void          _pika_unit_set_deletion_flag            (Pika        *pika,
                                                       PikaUnit     unit,
                                                       gboolean     deletion_flag);

gdouble       _pika_unit_get_factor                   (Pika        *pika,
                                                       PikaUnit     unit);

gint          _pika_unit_get_digits                   (Pika        *pika,
                                                       PikaUnit     unit);

const gchar * _pika_unit_get_identifier               (Pika        *pika,
                                                       PikaUnit     unit);

const gchar * _pika_unit_get_symbol                   (Pika        *pika,
                                                       PikaUnit     unit);
const gchar * _pika_unit_get_abbreviation             (Pika        *pika,
                                                       PikaUnit     unit);
const gchar * _pika_unit_get_singular                 (Pika        *pika,
                                                       PikaUnit     unit);
const gchar * _pika_unit_get_plural                   (Pika        *pika,
                                                       PikaUnit     unit);

void           pika_user_units_free                   (Pika        *pika);


#endif  /*  __APP_PIKA_UNIT_H__  */
